/*
 * Copyright (C) 1997  Warren Jasper
 * All rights reserved.
 *
 * This software may be freely copied, modified, and redistributed
 * provided that this copyright notice is preserved on all copies.
 *
 * You may not distribute this software, in whole or in part, as part of
 * any commercial product without the express consent of the authors.
 *
 * There is no warranty or other guarantee of fitness of this software
 * for any purpose.  It is provided solely "as is".
 *
 *
 *   Note: Interrupts, when enabled, are generated differently depending
 *   upon the mode of the board:
 *
 *   Soft Trigger:   poll IRQDATA for A/D done.
 *   Pacer clock:    interrupts generated with FIFO half full
 *                   or when Total_Count = 0.
 */

/***************************************************************************
 *
 * a2dc.c
 *
 ***************************************************************************/

#define MODULE

/* retrieve the CONFIG_* macros */
#include <linux/autoconf.h>
#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
#define MODVERSIONS
#endif

#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/version.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/major.h>
#include <linux/mm.h>
#include <linux/timer.h>
#include <linux/ioport.h>
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/signal.h>
#include <linux/config.h>       /* for CONFIG_PCI */
#include <linux/delay.h>
#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>
#include "pci-das1602.h"
#include "a2dc.h"

#ifdef CONFIG_PCI
#include <linux/pci.h>
#include <linux/bios32.h>
#endif


/***************************************************************************
 *
 * Prototype of pubic and private functions.
 *
 ***************************************************************************/

static int  SoftRead(WORD *outBuff, ADC_ChanRec *chan);
static int  PacerRead(WORD *outBuff, ADC_ChanRec *chan);
static int  ExternRead(WORD *outBuff, ADC_ChanRec *chan);
static int  SetADCPacerFreq(BoardRec *board);
static int  SetDACPacerFreq(BoardRec *board);
static int  SetADCChannelMux(BYTE lowChan, BYTE highChan);
static int  SetADCGain(WORD gain);
static void StartADCPacer();
static void StartDACPacer();
static void StopADCPacer();
static void StopDACPacer();
static void LoadADCPacer(BoardRec *board);
static void LoadDACPacer(BoardRec *board);
static void LoadCounter0(LONG arg);
static void LoadResidualCount(WORD resCount);

static void pci_das1602_AD_HalfFullInt();
static void pci_das1602_AD_PretrigInt();
static void pci_das1602_AD_NotEmptyInt();
static void pci_das1602_DA_FIFOEmptyInt();
static void pci_das1602_DA_HalfFullInt();

static int das1602_read(struct inode *iNode, struct file *filePtr, char *buf, int count);
static int das1602_write(struct inode *iNode, struct file *filePtr, const char *buf, int count);
static int das1602_open(struct inode *iNode, struct file *filePtr);
static void das1602_close(struct inode *iNode, struct file *filePtr);
static int das1602_ioctl(struct inode *iNode, struct file *filePtr, unsigned int cmd, LONG arg);
static void das1602_ReadInterrupt(int irq, void *dev_id, struct pt_regs *regs);
static void das1602_TimerHandler(LONG);
static void pci_das1602_aout( int channel, int gain, WORD value );
static void pci_das1602_aout_scan( int channel, int gain, WORD *DACbuf, int count );
static void calibrate_pci_das1602_aout( int channel, int gain );
static void calibrate_pci_das1602_ain( int gain );
static BYTE get_pci_das1602_nVRam( WORD addr );
static void write_pci_das1602_8800 ( BYTE addr, BYTE value );
static void write_pci_das1602_8402 ( BYTE addr, BYTE value );
static void write_pci_das1602_dac08 ( BYTE value );
static int config_region( unsigned int pci_ioaddr, int size, WORD *base );

int init_module(void);
void cleanup_module(void) ;


/***************************************************************************
 *
 * Global data.
 *
 ***************************************************************************/

static WORD base0;                          /* base register 0 address */
static WORD base1;                          /* base register 1 address */
static WORD base2;                          /* base register 2 address */
static WORD base3;                          /* base register 3 address */
static WORD base4;                          /* base register 4 address */
static int *base_address;                   /* pointer to address      */
static int WordsToRead;                     /* number of conversions until done  */
static BYTE dio_reg = 0x80;                 /* DIO control register              */
static ADC_ChanRec *CurrChan = NULL;        /* pointer to ChanADC[minor]         */
static int MajorNumber = DEFAULT_MAJOR_DEV; /* Major number compiled in          */
static ADC_ChanRec ChanADC[AD_CHANNELS];    /* ADC Channel specific information  */
static DAC_ChanRec ChanDAC[DA_CHANNELS];    /* DAC Channel specific information  */
static DIO_ChanRec ChanDIO[DIO_PORTS];      /* DIO Channel specific information  */
static BoardRec BoardData;                  /* Board specific information        */
static WORD KernBuff[MAX_COUNT];            /* Kernel buffer where samples are   */
                                            /* saved before write to user space  */
static WORD* KernBuffPtr;                   /* pointer to kernbuf                */
struct wait_queue  *adc0_wait;              /* wait semaphore                    */
struct timer_list TimerList = {NULL, NULL, 0, 0, das1602_TimerHandler};

/***************************************************************************
 *
 *
 ***************************************************************************/

static struct file_operations das1602_fops = {
	NULL,            /* seek    */
	das1602_read,    /* read    */
	das1602_write,   /* write   */
	NULL,		 /* readdir */
	NULL,		 /* select  */
	das1602_ioctl,	 /* ioctl   */
	NULL,
	das1602_open,
	das1602_close,
	NULL		 /* fsync   */
};

#ifdef __cplusplus
extern "C" {
#endif


/***************************************************************************
 *
 * Block copy of data samples from data register to kernel memory.
 *
 ***************************************************************************/

inline char *RegCopyToKernel(char *dest, const int reg, int n)
{
__asm__ __volatile__(
	"cld\n\t"
        "rep\n\t"
        "insw"
	: 
	:"D" (dest), "d" (reg),"c" (n)
	:"di","dx","cx");
return dest;
}

/***************************************************************************
 *
 * Block copy of data samples to data register from kernel memory.
 *
 ***************************************************************************/

inline char *RegCopyFromKernel(const int reg, char *dest, int n)
{
__asm__ __volatile__(
	"cld\n\t"
        "rep\n\t"
        "outsw"
	: 
	:"d" (reg), "D" (dest), "c" (n)
	:"dx","di","cx");
return dest;
}

/***************************************************************************
 *
 * Loads driver. Called when "insmod das16-M1.o" is invoked on the command line.
 *               The board is set to IRQ, 
 *
 ***************************************************************************/

int init_module(void) 
{
  int pci_index;
  unsigned char pci_bus, pci_device_fn;
  unsigned int pci_ioaddr;
  unsigned char pci_irq_line;
  int i, err;
  WORD wReg;

#ifndef CONFIG_PCI
    /* this should not happen. */
    printk("%s pci not configured with kernel!\n", ADAPTER_ID);
    return -ENODEV;
#endif

  /* Register as a device with kernel.  */

  if ((err = register_chrdev(MajorNumber, "pci-das1602/16", &das1602_fops))) {
      printk("%s: Failure to load module. error %d\n", ADAPTER_ID, -err);
      return err;
  } 

  /* Register the region of IO space with kernel and get base address */

  if(pcibios_present()) {
    for ( pci_index = 0; pci_index < MAX_BOARDS; pci_index++ ) {
      if ( pcibios_find_device (PCI_VENDOR_ID_CBOARDS,
              PCI_DEVICE_ID_CBOARDS_DAS1602_16, pci_index, &pci_bus, 
              &pci_device_fn) != 0 ) break;

      /* Get PCI_BASE_ADDRESS_0 */
      pcibios_read_config_dword( pci_bus, pci_device_fn,
                            PCI_BASE_ADDRESS_0, &pci_ioaddr );
      /* Strip the I/O address out of the returned value */
      pci_ioaddr &= PCI_BASE_ADDRESS_IO_MASK;
      if (config_region( pci_ioaddr, BADR0_SIZE, &base0 ) != 0 ) {
        printk("%s: Can't allocate region port address %#x\n", 
                ADAPTER_ID, pci_ioaddr );
        if (unregister_chrdev(MajorNumber, "pci-das1602/16") != 0) {
          printk("%s: unregister_chrdev() failed.\n", ADAPTER_ID);
        }
        return -ENODEV;
      }
      if ( pci_ioaddr != base0 ) {
          pci_ioaddr = base0 | 0x1;
          pcibios_write_config_dword( pci_bus, pci_device_fn,
                        PCI_BASE_ADDRESS_0, pci_ioaddr );
      }

      /* Get PCI_BASE_ADDRESS_1 */
      pcibios_read_config_dword( pci_bus, pci_device_fn,
                            PCI_BASE_ADDRESS_1, &pci_ioaddr );
      /* Strip the I/O address out of the returned value */
      pci_ioaddr &= PCI_BASE_ADDRESS_IO_MASK;
      if (config_region( pci_ioaddr, BADRn_SIZE, &base1 ) != 0 ) {
        printk("%s: Can't allocate region port address %#x\n", 
                ADAPTER_ID, pci_ioaddr );
        if (unregister_chrdev(MajorNumber, "pci-das1602/16") != 0) {
          printk("%s: unregister_chrdev() failed.\n", ADAPTER_ID);
        }
        return -ENODEV;
      }
      if ( pci_ioaddr != base1 ) {
          pci_ioaddr = base1 | 0x1;
          pcibios_write_config_dword( pci_bus, pci_device_fn,
                        PCI_BASE_ADDRESS_1, pci_ioaddr );
      }

      /* Get PCI_BASE_ADDRESS_2 */
      pcibios_read_config_dword( pci_bus, pci_device_fn,
                                 PCI_BASE_ADDRESS_2, &pci_ioaddr );
      /* Strip the I/O address out of the returned value */
      pci_ioaddr &= PCI_BASE_ADDRESS_IO_MASK;
      if (config_region( pci_ioaddr, BADRn_SIZE, &base2 ) != 0 ) {
        printk("%s: Can't allocate region port address %#x\n", 
                ADAPTER_ID, pci_ioaddr );
        if (unregister_chrdev(MajorNumber, "pci-das1602/16") != 0) {
          printk("%s: unregister_chrdev() failed.\n", ADAPTER_ID);
        }
        return -ENODEV;
      }
      if ( pci_ioaddr != base2 ) {
          pci_ioaddr = base2 | 0x1;
          pcibios_write_config_dword( pci_bus, pci_device_fn,
                       PCI_BASE_ADDRESS_2, pci_ioaddr );
      }

      /* Get PCI_BASE_ADDRESS_3 */
      pcibios_read_config_dword( pci_bus, pci_device_fn,
                              PCI_BASE_ADDRESS_3, &pci_ioaddr );
      /* Strip the I/O address out of the returned value */
      pci_ioaddr &= PCI_BASE_ADDRESS_IO_MASK;
      if (config_region( pci_ioaddr, BADRn_SIZE, &base3 ) != 0 ) {
        printk("%s: Can't allocate region port address %#x\n", 
                ADAPTER_ID, pci_ioaddr );
        if (unregister_chrdev(MajorNumber, "pci-das1602/16") != 0) {
          printk("%s: unregister_chrdev() failed.\n", ADAPTER_ID);
        }
        return -ENODEV;
      }
      if ( pci_ioaddr != base3 ) {
          pci_ioaddr = base3 | 0x1;
          pcibios_write_config_dword( pci_bus, pci_device_fn,
                        PCI_BASE_ADDRESS_3, pci_ioaddr );
      }

      /* Get PCI_BASE_ADDRESS_4 */
      pcibios_read_config_dword( pci_bus, pci_device_fn,
                           PCI_BASE_ADDRESS_4, &pci_ioaddr );
      /* Strip the I/O address out of the returned value */
      pci_ioaddr &= PCI_BASE_ADDRESS_IO_MASK;
      if (config_region( pci_ioaddr, BADRn_SIZE, &base4 ) != 0 ) {
        printk("%s: Can't allocate region port address %#x\n", 
                ADAPTER_ID, pci_ioaddr );
        if (unregister_chrdev(MajorNumber, "pci-das1602/16") != 0) {
          printk("%s: unregister_chrdev() failed.\n", ADAPTER_ID);
        }
        return -ENODEV;
      }
      if ( pci_ioaddr != base4 ) {
          pci_ioaddr = base4 | 0x1;
          pcibios_write_config_dword( pci_bus, pci_device_fn,
                        PCI_BASE_ADDRESS_4, pci_ioaddr );
      }

      base_address = (void *) base0;

      /* Register interrupt handler. */

      pcibios_read_config_byte( pci_bus, pci_device_fn, PCI_INTERRUPT_LINE, &pci_irq_line );
      BoardData.irq = pci_irq_line;
      if (request_irq(pci_irq_line, das1602_ReadInterrupt, (SA_INTERRUPT | SA_SHIRQ), 
          "pci-das1602/16", base_address)) {
        /* No free irq found! cleanup and exit */
        printk("%s: Can't request IRQ %d\n", ADAPTER_ID, BoardData.irq);
        if (unregister_chrdev(MajorNumber, "pci-das1602/16") != 0) { 
          printk("%s: unregister_chrdev() failed.\n", ADAPTER_ID);
	}
        release_region( base0,  BADR0_SIZE );       /* release region */
        release_region( base1,  BADRn_SIZE );       /* release region */
        release_region( base2,  BADRn_SIZE );       /* release region */
        release_region( base3,  BADRn_SIZE );       /* release region */
        release_region( base4,  BADRn_SIZE );       /* release region */
        return -EADDRNOTAVAIL;
        break;
      }
    }

    if ( pci_index == 0 ) {
      printk("%s No device present.\n", ADAPTER_ID);
      return -ENODEV;
    }
  } else {
    printk("%s No pci bios present.\n", ADAPTER_ID);
    return -ENODEV;
  }
  BoardData.base0 = base0;
  BoardData.base1 = base1;
  BoardData.base2 = base2;
  BoardData.base3 = base3;
  BoardData.base4 = base4;

  /* Set all channel structures to show nothing active/open */
  for (i = 0; i < AD_CHANNELS; i++) {
      ChanADC[i].open = FALSE;
      ChanADC[i].lowChan = i;
      ChanADC[i].hiChan = i;
      ChanADC[i].gain = UP_10_00V;   /* set range to 0-10V */
                                     /* Differential Mode  */
      ChanADC[i].nSpare2 = 0;
  }

  for ( i = 0; i < DA_CHANNELS; i++ ) {
      ChanDAC[i].open = FALSE;
      ChanDAC[i].gain = UP_10_00V;   /* set range to 0-10V */
      ChanDAC[i].nSpare3 = 0;
  }

  ChanDIO[0].addr = DIO_PORTA;
  ChanDIO[1].addr = DIO_PORTB;
  ChanDIO[2].addr = DIO_PORTC;

  for ( i = 0; i < DIO_PORTS; i++ ) {
    ChanDIO[i].open = FALSE; 
    ChanDIO[i].mode = 0;
    outb_p(0x0, ChanDIO[i].addr);
  }

  adc0_wait = NULL;

  /* Reset FIFO interrupts and disable all interupts */
  wReg = (ADFLCL | INTCL | EOACL | DAHFCL | DAEMCL );
  BoardData.nSpare0 = wReg;
  outw_p( wReg, IRQ_REG);

  /* Clear TRIG_REG register */
  outw_p(0x0, TRIG_REG);
  BoardData.nSpare2 = 0x0;

  /* Clear DAC empty interrupt.  Select unipolar 10V
     range for both DACS and zero them out by calling
     pci1602_AOut.  Note that nSpare3 is made 0 so that
     pci1602_AOut is forced to load cal. coeffs
     PTR1 BASE+8
  */

  BoardData.nSpare3 = 0x0;
  pci_das1602_aout( 0, UP_10_00V, 0x0 );
  pci_das1602_aout( 1, UP_10_00V, 0x0 );

  /* Clear DAC FIFO */
  outw(0x1, ADC_FIFO_CLR);

  /* Load the ADC calibration coeffs for unipolar 10V */
  calibrate_pci_das1602_ain( UP_10_00V );

  BoardData.ADC_freq = DEFAULT_FREQ;   /* Set default pacer clock frequency */
  SetADCPacerFreq(&BoardData);

  /* Reset Interrupt on AMCC5933 PCI controller */
  outl(INTCSR_DWORD, INTCSR_ADDR);
  outl(BMCSR_DWORD, BMCSR_ADDR);

  BoardData.busy = FALSE;          /* board ready */

  printk("%s: BADR0=%#x BADR1=%#x BADR2=%#x BADR3=%#x BADR4=%#x\n",
          ADAPTER_ID, base0, base1, base2, base3, base4);
  printk("%s: IRQ=%d 10/3/97 wjasper@tx.ncsu.edu\n", ADAPTER_ID, BoardData.irq);

  return 0;
}

/***************************************************************************
 *
 * Test memory locations for exsistance of pci-das1602/16 board.
 *
 ***************************************************************************/
static int config_region( unsigned int pci_ioaddr, int size, WORD *base )
{

  /* check to see if the allocated region is ok */
  if ( check_region(pci_ioaddr, size) == 0 ) {
    request_region( pci_ioaddr, size, "pci-das1602/16" );
    *base = pci_ioaddr;
    return 0;
  } else {
    /* Look to see if there is another place in the io space */
    for ( pci_ioaddr = PORT_MIN; pci_ioaddr < PORT_MAX; pci_ioaddr += PORT_STEP ) {
      if ( check_region(pci_ioaddr,  size ) == 0 ) {
        request_region( pci_ioaddr, size, "pci-das1602/16" );
        *base = pci_ioaddr;
        return 0;
      }
    }
    
    if ( pci_ioaddr > PORT_MAX ) {
      printk("%s: No free address from %#x to %#x.\n", 
             ADAPTER_ID, PORT_MIN, PORT_MAX);
      if (unregister_chrdev(MajorNumber, "pci-das1602/16") != 0) {
          printk("%s: unregister_chrdev() failed.\n", ADAPTER_ID);
      }
      return -ENODEV;
    }
  }
  return 1;
}

/***************************************************************************
 *
 * Remove driver. Called when "rmmod pci-das1602/16" is run on the command line.
 *
 ***************************************************************************/

void cleanup_module(void) 
{
  if (MOD_IN_USE) {
    printk("%s: device busy, remove delayed.\n", ADAPTER_ID);
    return;
  }

  StopADCPacer();
  StopDACPacer();
  release_region( base0,  BADR0_SIZE );       /* release region */
  release_region( base1,  BADRn_SIZE );       /* release region */
  release_region( base2,  BADRn_SIZE );       /* release region */
  release_region( base3,  BADRn_SIZE );       /* release region */
  release_region( base4,  BADRn_SIZE );       /* release region */
  
  free_irq(BoardData.irq, base_address);

  if (unregister_chrdev(MajorNumber, "pci-das1602/16") != 0) {
    printk("%s: cleanup_module failed.\n", ADAPTER_ID);
  } else {

  #ifdef DEBUG
      printk("%s: module removed.\n", ADAPTER_ID);
  #endif

  }    
}

#ifdef __cplusplus
}
#endif

/***************************************************************************
 *
 * open() service handler
 *
 ***************************************************************************/

static int das1602_open(struct inode *iNode, struct file *filePtr)
{
  int minor = MINOR(iNode->i_rdev);
  int port;
  WORD wReg;

  /* 
     check if device is already open: only one process may read from a
     port at a time.  There is still the possibility of two processes 
     reading from two different channels messing things up. However,
     the overhead to check for this may not be worth it.
  */

  if ( minor >= 0 && minor < AD_CHANNELS ) {
    if ( ChanADC[minor].open == TRUE ) {
      return -EBUSY;
    }

    MOD_INC_USE_COUNT;
    ChanADC[minor].open = TRUE;              /* The device is open */
    ChanADC[minor].gain = BP_10_00V;         /* +/- 10V , Differential Mode */
    ChanADC[minor].mode = filePtr->f_flags;  /* set acquisition mode */
    ChanADC[minor].nSpare2 = 0;

    /* Reset FIFO interrupts and disable all interupts */
    wReg = (ADFLCL | INTCL | EOACL | DAHFCL | DAEMCL );
    BoardData.nSpare0 = wReg;
    outw_p(wReg, IRQ_REG);

    /* Clear TRIG_REG register */
    outw_p(0x0, TRIG_REG);
    BoardData.nSpare2 = 0x0;

    #ifdef DEBUG
        printk("%s: open(): minor %d mode %d.\n", ADAPTER_ID, minor, ChanADC[minor].mode);
    #endif
    return 0;   
  }

  if ( (minor >= AD_CHANNELS) && (minor < AD_CHANNELS+DIO_PORTS) ) {
    port = minor - AD_CHANNELS;
    if ( ChanDIO[port].open == TRUE ) {
      return -EBUSY;
    }

    MOD_INC_USE_COUNT;
    ChanDIO[port].open = TRUE;                 /* The device is open */
    ChanDIO[port].f_flags = filePtr->f_flags;

    #ifdef DEBUG
      printk("%s: open(): minor %d mode %d.\n", ADAPTER_ID, minor, ChanDIO[port].mode);
    #endif
    return 0;
  }

  if ( (minor >= AD_CHANNELS+DIO_PORTS) && (minor < AD_CHANNELS+DIO_PORTS+DA_CHANNELS) ) {
    port = minor - AD_CHANNELS - DIO_PORTS;
    if ( ChanDAC[port].open == TRUE ) {
      return -EBUSY;
    }

    MOD_INC_USE_COUNT;
    ChanDAC[port].open = TRUE;                 /* The device is open */
    ChanDAC[port].mode = filePtr->f_flags;
    #ifdef DEBUG
      printk("%s: open(): minor %d mode %d.\n", ADAPTER_ID, minor, ChanDIO[port].mode);
    #endif

  }
  return 0;
}

/***************************************************************************
 *
 * close() service handler
 *
 ***************************************************************************/

static void das1602_close(struct inode *iNode, struct file *filePtr)
{
  int minor = MINOR(iNode->i_rdev);
  int port;


  MOD_DEC_USE_COUNT;

  if ( minor >= 0 && minor < AD_CHANNELS ) {
    ChanADC[minor].open = FALSE;
  } else if ( minor >= AD_CHANNELS && minor < AD_CHANNELS + DIO_PORTS ) {
    port = minor - AD_CHANNELS;
    ChanDIO[port].open = FALSE;
  } else if ( minor >= AD_CHANNELS+DIO_PORTS && minor < AD_CHANNELS+DIO_PORTS+DA_CHANNELS ) {
    port = minor - AD_CHANNELS - DIO_PORTS;
    ChanDAC[port].open = FALSE;
  } else {
    printk("das1602_close: Incorrect minor number (%d).\n", minor);
  }  

  #ifdef DEBUG
      printk("%s: close() of minor number %d.\n", ADAPTER_ID, minor);
  #endif
}

/***************************************************************************
 *
 * read() service function
 *
 ***************************************************************************/

static int das1602_read(struct inode *iNode, struct file *filePtr, char *buf, int count)
{
  int   stat;
  BYTE  bReg;
  int   minor = MINOR(iNode->i_rdev);
  int   port;

  if ( BoardData.busy == TRUE ) {   /* if board is in use, return busy */
      return (-EBUSY);
  } else {
      BoardData.busy = TRUE;
  }

  if ( count < 1 ) {
    printk("das1602_read(): count must be greater than 0.\n");
    return (-1);
  }

  if ( minor >= 0 && minor < AD_CHANNELS ) {
    if ( (ChanADC[minor].count = count) > MAX_COUNT ) {
        printk("das1602_read(): requesting too large a count size.\n");
        BoardData.busy = FALSE;
        return (-1);
    }

    /* Read */

    switch (ChanADC[minor].mode) {

      case ADC_SOFT_CONVERT:
          #ifdef DEBUG
              printk("das1602_read(): Entering ADC_SOFT_TRIGGER mode.  count = %d\n",
                      count);
          #endif

          if (SoftRead(KernBuff, &ChanADC[minor])) {
              printk("das1602_read: SoftTrigRead() failed.\n");
              BoardData.busy = FALSE;
              return(-1);
          } 
          break;

      case ADC_EXTERNAL_PACER_FALLING:
      case ADC_EXTERNAL_PACER_RISING:

          #ifdef DEBUG
              printk("das1602_read(): Entering ADC_EXTERNAL_PACER mode.\n");
          #endif

          if (ExternRead(KernBuff, &ChanADC[minor])) {
              printk("das1602_read: ExternalRead() failed with ADC_EXTERNAL_PACER.\n");
              BoardData.busy = FALSE;
              return(-1);
          }
          break;

      case ADC_PACER_CLOCK:

          #ifdef DEBUG
              printk("das1602_read(): Entering ADC_PACER_CLOCK mode.\n");
          #endif

          if ( ChanADC[minor].count == 1 ) {            /* use SoftTrigger */
              if (SoftRead(KernBuff, &ChanADC[minor])) {
                  printk("das1602_read: SoftRead() failed with pacer.\n");
                  BoardData.busy = FALSE;
                  return(-1);
              }
          } else {
              StopADCPacer();                        /* disable pacer if in pacer mode */
              if (PacerRead(KernBuff, &ChanADC[minor])) {
                  printk("das1602_read: PacerRead() failed.\n");
                  BoardData.busy = FALSE;
                  return(-1);
              } 
          }
          break;
    }                  /* end switch */

    /* Check for overflows */

    /* Check that data can be written to file. */
    if ((stat = verify_area(VERIFY_WRITE, buf, sizeof(WORD)*ChanADC[minor].count)) != 0) {
      printk("das1602_read: Failed VERIFY_WRITE.\n");
      BoardData.busy = FALSE;
      return -1;
    }

    /* Write data to user space */
    if (ChanADC[minor].count == 1) {
      put_fs_word(*KernBuff, (WORD*) buf);
    } else  {
      memcpy_tofs(buf, KernBuff, ChanADC[minor].count*sizeof(WORD));
    }

    BoardData.busy = FALSE;
    return(ChanADC[minor].count);     /* return number of samples read */
  }

  /* check to see if reading a value from the DIO */

  if ( minor >= AD_CHANNELS && minor < AD_CHANNELS + DIO_PORTS ) {
    port = minor - AD_CHANNELS;
    bReg = inb(ChanDIO[port].addr);
    put_fs_byte(bReg, buf);

    #ifdef DEBUG
       printk("das1602_read: %s DIO Port %#x addr %#x set to %#x\n", 
                             ADAPTER_ID, port, ChanDIO[port].addr, bReg);
    #endif
    BoardData.busy = FALSE;
    return 1;
  }

  printk("das1602_read: Incorrect minor number (%d).\n", minor);
  BoardData.busy = FALSE;
  return -1;
}

/***************************************************************************
 *
 * write() service function
 *
 ***************************************************************************/

static int das1602_write(struct inode *iNode, struct file *filePtr, const char *buf, int count)
{
  int  minor = MINOR(iNode->i_rdev);
  int  port;
  BYTE bReg;
  WORD DACbuf[FIFO_SIZE];
  int stat;
  register int i;

  #ifdef DEBUG
    printk("das1602_write(): Minor = %d, Count = %d\n", minor, count);
  #endif

  if ( BoardData.busy == TRUE ) {   /* if board is in use, return busy */
      return (-EBUSY);
  } else {
      BoardData.busy = TRUE;
  }

  if ( count < 1 ) {
    printk("das1602_write(): count must be greater than 0.\n");
    return (-1);
  }

  /* check to see if writing a value to the DIO */
  if ( (minor >= AD_CHANNELS) && (minor < AD_CHANNELS+DIO_PORTS) ) {
    port = minor - AD_CHANNELS;
    bReg = get_fs_byte(buf);
    ChanDIO[port].value = bReg;
    outb( bReg, ChanDIO[port].addr);
    #ifdef DEBUG
    printk("das1602_write %s: DIO Port %#x set to %#x\n", ADAPTER_ID, port, bReg);
    #endif
    BoardData.busy = FALSE;
    return 1;
  } 

  if ( (minor >= AD_CHANNELS+DIO_PORTS) && (minor < AD_CHANNELS+DIO_PORTS+DA_CHANNELS) ) {
    port = minor - AD_CHANNELS - DIO_PORTS;
    ChanDAC[port].count = count;

    /* Check that data can be read from user space. */
    if ((stat = verify_area(VERIFY_READ, buf, sizeof(WORD)*ChanDAC[port].count)) != 0) {
      printk("das1602_read: Failed VERIFY_READ.\n");
      BoardData.busy = FALSE;
      return -1;
    }

    /* Read data from user space */
    if (ChanDAC[port].count == 1) {
      DACbuf[0] = get_fs_word(buf);
    } else  {
      memcpy_fromfs(DACbuf, buf, count*sizeof(WORD));
    }

    switch (ChanDAC[port].mode) {

      case DAC_SOFT_CONVERT:
        #ifdef DEBUG
            printk("das1602_write(): Entering DAC_SOFT_CONVERT mode. port = %d, gain = %d\n",
                    port, ChanDAC[port].gain );
        #endif
        for ( i = 0; i < count; i++ ) {
            pci_das1602_aout( port, ChanDAC[port].gain, DACbuf[i] );
	}
        break;

      case DAC_PACER_CLOCK:
        #ifdef DEBUG
            printk("das1602_write(): Entering DAC_PACER_CLOCK mode.\n");
        #endif
        break;

      case DAC_FALLING_EDGE:
        #ifdef DEBUG
            printk("das1602_write(): Entering DAC_FALLING_EDGE mode.\n");
        #endif
        break;

      case DAC_RISING_EDGE:
        #ifdef DEBUG
            printk("das1602_write(): Entering DAC_RISING_EDGE mode.\n");
        #endif
        break;
    }
  }
  BoardData.busy = FALSE;
  return(ChanDAC[minor].count);     /* return number of samples written */


}

/***************************************************************************
 *
 * iotctl() service handler
 *
 ***************************************************************************/

/* Note:
    Remember that FIOCLEX, FIONCLEX, FIONBIO, and FIOASYN are reserved ioctl cmd numbers
*/

static int das1602_ioctl(struct inode *iNode, struct file *filePtr, unsigned int cmd, LONG arg)
{
  int minor = MINOR(iNode->i_rdev);
  int port;
  int channel;

  if ( minor >= 0 && minor < AD_CHANNELS ) {
    switch (cmd) {
      case ADC_SET_GAINS:
        #ifdef DEBUG
          printk("ioctl ADC_SET_GAINS: Channel = %d, Gain = %ld\n", minor, arg);
        #endif
        ChanADC[minor].gain &= ~(UNIBIP | GS1 | GS0);
        ChanADC[minor].gain |=  (WORD) arg;
        break;
      case ADC_GET_GAINS:
        put_fs_long((long)ChanADC[minor].gain, (long*) arg);
        break;
      case ADC_SET_PACER_FREQ:
        if ( arg > MAX_AD_FREQ ) {
            printk("ioctl:  Can not set frequency %ld greater than %d.\n", arg, MAX_AD_FREQ );
            return -1;
        } else {
            BoardData.ADC_freq = (LONG)arg;
            SetADCPacerFreq(&BoardData);
            LoadADCPacer(&BoardData);               /* load the board frequency now */
        }
        break;
      case ADC_GET_PACER_FREQ:
        put_fs_long(BoardData.ADC_freq, (long*) arg);
        break;
      case ADC_START_PACER:
        StartADCPacer();
        break;
      case ADC_STOP_PACER:
        StopADCPacer();
        break;
      case ADC_COUNTER0:
        LoadCounter0( (LONG) arg );
        break;
      case ADC_SET_MUX_LOW:
        ChanADC[minor].lowChan = (BYTE) arg;
        break;
      case ADC_SET_MUX_HIGH:
        ChanADC[minor].hiChan = (BYTE) arg;
        break;
      case ADC_GET_CHAN_MUX_REG:
        put_fs_long( BoardData.nSpare1, (long*) arg);
        break;
      case ADC_SET_FRONT_END:
        if ( arg ) {
          ChanADC[minor].gain |= SEDIFF;     /* Single Ended Mode */
        } else {
          ChanADC[minor].gain &= ~(SEDIFF);  /* Differential Mode */
	}
        break;
      case ADC_BURST_MODE:
        if (arg == 0) {
          ChanADC[minor].nSpare2 &= ~BURSTE;
	} else {
          ChanADC[minor].nSpare2 |= BURSTE;
	}
        break;
      default:
        return(-EINVAL);
        break;
    } /* end switch */
  }  /* end if */

  if ( minor >= AD_CHANNELS && minor < AD_CHANNELS+DIO_PORTS ) {
    port = minor - AD_CHANNELS;
    switch (cmd) {
      case DIO_SET_MODE:
        arg &= 0x3;
        switch (port) {
	    case 0:		       /* Port A */
              #ifdef DEBUG
              printk("DIO_SET_MODE for Port A\n");
              #endif
              dio_reg &= 0x9f;
              dio_reg |= (arg << 5);
              outb_p(dio_reg, DIO_CNTRL_REG);
              ChanDIO[port].mode = arg;
              break;

            case 1:			/* Port 1B */
              #ifdef DEBUG
              printk("DIO_SET_MODE for Port 1B\n");
              #endif
              if (arg & 0x2)
	        return(-EINVAL);	       /* Port 1B only has Modes 0 & 1 */
              dio_reg &=  0xfb;
              dio_reg |= (arg << 2);
              outb_p(dio_reg, DIO_CNTRL_REG);
              ChanDIO[port].mode = arg;
              break;

            case 2:			/* Port 1C */
              #ifdef DEBUG
              printk("DIO_SET_MODE for Port 1C\n");
              #endif
              if (arg)
	        return(-EINVAL);	/* Port 1C only has Mode 0 I/O */
              ChanDIO[port].mode = arg;
              break;

            default:
              #ifdef DEBUG
              printk("DIO_SET_MODE for Invalid Port\n");
              #endif
              return(-EINVAL);	/* Wrong Port Number */
              break;
        }
        break;
    
      case DIO_SET_DIRECTION:

        #ifdef DEBUG
        printk("DIO_SET_DIRECTION: minor = %d, arg = %d\n", (int) minor, (int) arg );
        #endif

        switch (port) {
          case 0:			/* Port A */
            #ifdef DEBUG
            printk("DIO_SET_DIRECTION for Port A\n");
            #endif
            arg &= 0x1;
            dio_reg &=  0xef;
            dio_reg |= (arg << 4);
            outb_p(dio_reg, DIO_CNTRL_REG);
            break;

          case 1:			/* Port B */
            #ifdef DEBUG
            printk("DIO_SET_DIRECTION for Port B\n");
            #endif
            arg &= 0x1;
            dio_reg &=  0xfd;
            dio_reg |= (arg << 1);
            outb_p(dio_reg, DIO_CNTRL_REG);
            break;

          case 2:			/* Port C */
            #ifdef DEBUG
            printk("DIO_SET_DIRECTION for Port C\n");
            #endif
            switch ( arg ) {
              case PORT_INPUT:         /* CU = IN  CL = IN */
                  dio_reg |= 0x9;
                  break;
              case PORT_OUTPUT:        /* CU = OUT  CL = OUT */
                  dio_reg &= ~(0x9);
                  break;
              case LOW_PORT_INPUT:     /* CL = IN */
                  dio_reg |= 0x1;
                  break;
              case LOW_PORT_OUTPUT:    /* CL = OUT */
                  dio_reg &= ~(0x1);
                  break;
              case HIGH_PORT_INPUT:    /* CU = IN */
                  dio_reg |= 0x8;
                  break;
              case HIGH_PORT_OUTPUT:   /* CU = OUT */
                  dio_reg &= ~(0x8);
                  break;
              default:
                  return(-EINVAL);
                  break;
            }
            outb_p(dio_reg, DIO_CNTRL_REG);
            break;

          default:
            #ifdef DEBUG
            printk("DIO_SET_DIRECTION for Invalid Port\n");
            #endif
            return(-EINVAL);
            break;
        }
        break;
    
      default:
        return(-EINVAL);
        break;
    }  /* end switch */
  }    /* end if */

  if ( minor >= AD_CHANNELS+DIO_PORTS && minor < AD_CHANNELS+DIO_PORTS+DA_CHANNELS ) {
    channel = minor - AD_CHANNELS - DIO_PORTS;
    switch (cmd) {
      case DAC_SET_GAINS:
        #ifdef DEBUG
          printk("ioctl DAC_SET_GAINS: Channel = %d, Gain = %ld\n", channel, arg);
        #endif
        ChanDAC[channel].gain = arg;
        break;
      case DAC_GET_GAINS:
        put_fs_long((long)ChanDAC[channel].gain, (long*) arg);
        break;
      case DAC_SET_PACER_FREQ:
        if ( arg > MAX_DA_FREQ ) {
          printk("ioctl:  Can not set frequency %ld greater than %d.\n", arg, MAX_DA_FREQ );
          return -1;
	} else {
          BoardData.DAC_freq = (LONG) arg;
          SetDACPacerFreq(&BoardData);
          LoadDACPacer(&BoardData);
	}
        break;
      case DAC_GET_PACER_FREQ:
        put_fs_long(BoardData.DAC_freq, (long*) arg);
        break;
      case DAC_START_PACER:
        StartADCPacer();
        break;
      case DAC_STOP_PACER:
        StopADCPacer();
        break;
    }
  }
  return 0;
}

/***************************************************************************
 *
 *  Alarm handler to handle situations where an interrupt is missed.
 *
 ***************************************************************************/

static void das1602_TimerHandler(LONG unused)
{

  /* 
     we should only be here if an interrupt occured while 
     we were reading half the FIFO 
  */

  #ifdef DEBUG
      printk("TimerHandler: WordsToRead = %d\n", WordsToRead);
  #endif

  cli();
 /* turn off interrupts, and trigger control.*/

  switch (CurrChan->mode) {
    case ADC_EXTERNAL_PACER_FALLING:
    case ADC_EXTERNAL_PACER_RISING:
      RegCopyToKernel((char *) KernBuffPtr, ADC_DATA_REG, WordsToRead+PACKETSIZE );
      break;
    case ADC_PACER_CLOCK:
      RegCopyToKernel((char *) KernBuffPtr, ADC_DATA_REG, WordsToRead );
      break;
  }
  WordsToRead = 0;
  wake_up_interruptible(&adc0_wait);  
  sti();
}

/***************************************************************************
 *
 * Interrupt handler used to service interrupt read().
 *
 ***************************************************************************/

static void das1602_ReadInterrupt(int irq, void *dev_id, struct pt_regs *regs)
{

  WORD IntSource;
  
  /* check to see if this is our interrupt */
  if (dev_id !=  base_address ) {
    return;
  }

  cli();
  
  if ( WordsToRead == 0 ) {
    sti();
    StopADCPacer();
    StopDACPacer();

    /* Reset Interrupt on AMCC5933 PCI controller */
    outl_p(INTCSR_DWORD, INTCSR_ADDR);
    outl_p(BMCSR_DWORD, BMCSR_ADDR);
    return;
  }

  /* Check the source of interrupt */
  IntSource = inw(IRQ_REG);

  #ifdef DEBUG
      printk("Entering das1602_ReadInterrupt(). irq = %d mode = %d  count = %d  IntSource = %#hx\n", 
              irq, CurrChan->mode, WordsToRead, IntSource );
  #endif

  while (IntSource) {
    if (IntSource & EOAI) {  /* End of acquisition */
      IntSource &= ~EOAI;
      pci_das1602_AD_HalfFullInt();
    } else if (IntSource & ADHFI) { /* AD-FIFO-half-full interrupt */
      IntSource &= ~ADHFI;
      if (BoardData.PreCount == 0) { 
       pci_das1602_AD_HalfFullInt();
      } else {
       pci_das1602_AD_PretrigInt();
      }
    } else if (IntSource & ADNEI) { /* AD-not-empty */
      IntSource &= ~ADNEI;
      pci_das1602_AD_NotEmptyInt();
    } else if (IntSource & EOBI) { /* end-of-burst interrrupt */
      IntSource &= ~EOBI;
      pci_das1602_AD_HalfFullInt();
    } else if (IntSource & DAEMI) { /* DA-FIFO-empty */
      IntSource &= ~DAEMI;
      pci_das1602_DA_FIFOEmptyInt();
    } else if (IntSource & DAHFI) { /* DA-FIFO-half-full interrupt */
      IntSource &= ~DAEMI;
      pci_das1602_DA_HalfFullInt();
    } else {
      break;
    }
  }

  if (WordsToRead == 0) {  /* We are done */
    /* turn off interrupts, and trigger control */
    /*    del_timer(&TimerList); */
    BoardData.nSpare0 &= ~(INTE | DAHFIE | DAEMIE | EOAIE);
    outw(BoardData.nSpare0, IRQ_REG);
    wake_up_interruptible(&adc0_wait);  
  }

  /* Reset Interrupt on AMCC5933 PCI controller */
  outl_p(INTCSR_DWORD, INTCSR_ADDR);
  outl_p(BMCSR_DWORD, BMCSR_ADDR);

  sti();
  return;
}

/***********************************************************************
 *  This routine gets called from das1602_ReadInterrupt() when the     *
 *  interrupt was caused by the FIFO going half full.  It reads half   *
 *  a FIFO's worth of data and stores it in the kernel's arrary.       *
 ***********************************************************************/

static void pci_das1602_AD_HalfFullInt()
{
  #ifdef DEBUG
      printk("Entering pci_das1602_AD_HalfFullInt()\n");
  #endif

  /* check for FIFO Overflow errors */
  if (inw_p(IRQ_REG) & LADFUL) {
    printk("pci_das1602_AD_HalfFullInt: ADC FIFO exceeded full state. Data may be lost\n");
    outw(0x1, ADC_FIFO_CLR);
    outw(BoardData.nSpare0 | ADFLCL | INTCL, IRQ_REG);  /* Clear Interrupts */
    return;
  }

  /* Caclulate remaining count.  If it's less than PACKET_SIZE then use
     remaining count, otherwise use PACKET_SIZE
  */

  if (WordsToRead  > FIFO_SIZE) {
    insw(ADC_DATA_REG, KernBuffPtr, PACKETSIZE);      /* uses rep insw */
    KernBuffPtr += PACKETSIZE;
    WordsToRead -= PACKETSIZE;
    if (WordsToRead <= 2*PACKETSIZE && WordsToRead > PACKETSIZE) {
      BoardData.nSpare2 |= ARM;             /* Arm the resisual at next ADHF */
      outw(BoardData.nSpare2,  TRIG_REG);   
    }
    printk("pci_das1602_AD_HalfFullInt > FIFO_SIZE. nSpare0 = %#x\n", BoardData.nSpare0);
  } else if (WordsToRead <= 2*PACKETSIZE && WordsToRead > PACKETSIZE) {
    if ( !(BoardData.nSpare2 & ARM)) {
      outw(BoardData.nSpare2 | ARM | FFM0, TRIG_REG);   /* Arm the resisual counter now */
    }
    insw(ADC_DATA_REG, KernBuffPtr, PACKETSIZE);        /* uses rep insw */
    KernBuffPtr += PACKETSIZE;
    WordsToRead -= PACKETSIZE;
    /* check for a race condition before leaving */
    if (inw(IRQ_REG) & EOAI) {
      insw(ADC_DATA_REG, KernBuffPtr, WordsToRead);       /* uses rep insw */
      KernBuffPtr += PACKETSIZE;
      WordsToRead =0;
    } else {
      printk("pci_das1602_AD_HalfFullInt < FIFO_SIZE. nSpare0 = %#x\n", BoardData.nSpare0);
    }
  } else {                                              /* we are done */
    printk("pci_das1602_AD_HalfFullInt < PACKET_SIZE. nSpare0 = %#x\n", BoardData.nSpare0);
    insw(ADC_DATA_REG, KernBuffPtr, WordsToRead);       /* uses rep insw */
    KernBuffPtr += PACKETSIZE;
    WordsToRead =0;
  }
  outw(BoardData.nSpare0 | INTCL, IRQ_REG);  /* Clear Interrupts */
}

static void pci_das1602_AD_PretrigInt()
{
  /* pretirggering not currently implemented */
  printk("pci_das1602_AD_PretrigInt: Error. Should not be called.\n");
}

/***********************************************************************
 *  This routine gets called from das1602_ReadInterrupt() when the     *
 *  interrupt was caused by any data going into the FIFO.  Usually     *
 *  gets called when sampling in SINGLEIO mode with interrupts.  For   *
 *  preformance reasons, I am polling in SoftTrigger Mode              *
 ***********************************************************************/

static void pci_das1602_AD_NotEmptyInt()
{

  #ifdef DEBUG
      printk("Entering pci_das1602_AD_NotEmptyInt()\n");
  #endif

  /* check for FIFO Overflow errors */
  if (inw(IRQ_REG) & LADFUL) {
    printk("pci_das1602_AD_NotEmptyInt: ADC FIFO exceeded full state. Data may be lost\n");
    return;
  }

  do {
    /* Read a sample from FIFO */
    *KernBuffPtr++ = inw(ADC_DATA_REG);
    WordsToRead--;
  } while ( WordsToRead && (inw(IRQ_REG) & ADNE) );

  outw(BoardData.nSpare0 | INTCL, IRQ_REG);  /* Clear Interrupts */
}

static void pci_das1602_DA_FIFOEmptyInt()
{
  printk("pci_das1602_DA_FIFOEmptyInt: Error. Should not be called.\n");
}

static void pci_das1602_DA_HalfFullInt()
{  
  printk("pci_das1602_DA_HalfFullInt: Error. Should not be called.\n");
}


/***************************************************************************
 *
 * Handles software/pacer triggered read().
 *
 * Bang! Bang! Method: 
 *    
 *    o Force interrupt by forcing conversion.
 *    o Get one sample at a time and put it in the kernel buffer(kernBuff).
 *    o Get chan->count samples
 *
 ***************************************************************************/

static int SoftRead(WORD *kernBuff, ADC_ChanRec *chanRec)
{
  WORD wReg;

  #ifdef DEBUG
      printk("Entering SoftRead().\n");
  #endif

  /* Disable everything first */
  outw(0x0, IRQ_REG);
  outw(0x0, TRIG_REG);

  /* Clear the AD FIFO */
  outw(0x1, ADC_FIFO_CLR);

  /* Set up the ADC Mux and Control register */
  SetADCChannelMux(chanRec->lowChan, chanRec->hiChan);  /* Read only this channel */
  SetADCGain(chanRec->gain);
  BoardData.nSpare1 &= ~(ADPS1 | ADPS0);
  outw_p(BoardData.nSpare1, MUX_REG);
  udelay(10);    /* wait 10 usecs for things to settle down */

  WordsToRead = chanRec->count;
  KernBuffPtr = KernBuff;

  /* Kludge:  I don't quite understand what is going on, but the
     very first conversion has garbage.  Just take it and move on.
  */

  outw_p(0x0, ADC_DATA_REG);  /* Force first conversion */
  do {
    wReg = inw_p(MUX_REG) & EOC;
  } while (!wReg);
 
  wReg  = (WORD) inw_p(ADC_DATA_REG); 

  while ( WordsToRead ) {
    outw_p(0x0, ADC_DATA_REG);  /* Force first conversion */

    /*
       Poll EOC for conversion.  Normally, I would use a interrupt
       handler, but the DAS1602 board is too fast, and we get
       a race condition.  Much better to poll for a single conversion
       here.
    */

    do {
      wReg = inw_p(MUX_REG) & EOC;
    } while (!wReg);

    *KernBuffPtr++ = (WORD) inw_p(ADC_DATA_REG);  /* Load into buffer */
    WordsToRead--;
  }

  #ifdef DEBUG
      printk("SoftRead(): value = %#x.\n", KernBuff[0]);
  #endif

  return 0;
}

/****************************************************************************
 *   Set ADC Pacer Clock Frequency
 * 
 *   Description: 
 *       Set the counters so that the ADC pacer clock runs at the
 *       desired frequency.  The frequency is generated by dividing
 *       down a 10 MHz clock, so all frequencies can not be generated.
 *       This routine calculated the divisor to generate a frequency
 *       as near as possible to the requested one.  It then calculates
 *       the real frequency and returns it .
 ****************************************************************************/

static int SetADCPacerFreq(BoardRec *board)
{
  WORD ctr1, ctr2;
  LONG product, error, error2;

  if ( board->ADC_freq == 0 ) {
    return -1;
  }

  /* divide 10MHz by frequency */
  product =  (long) 10000000 / board->ADC_freq;

  /* check for rounding error */
  error = abs(10000000 - product*board->ADC_freq);
  error2 = abs(10000000 - (product+1)*board->ADC_freq);
  if ( error2 < error ) product++;

  /* Now the job is to find two 16 bit numbers, that when multiplied
     together are approximately equal to product.  Start by setting
     one of them, ctr1 to 2 (minimum settable value) and increment until
     the error is minimized and ctr2 is less than 32768.

     NOTE: In Mode 2, a value of 1 is illegal! Therefore, crt1 and crt2
     can never be 1.

  */

  ctr1 = product / 32768;
  if ( ctr1 < 2 ) ctr1 = 2;
  ctr2 = product / ctr1;
  error = abs(product - (long) ctr2 * (long) ctr1);

  while ( error && ctr1 < 32768  && ctr2 > 1 ) {
       ctr1++;
       ctr2 = product / ctr1;
       error = abs(product - (long) ctr2 * (long) ctr1);
  }

  /* the frequency is prime, add 1 to it */
  if ( error ) {
      product++;
      ctr1 = product / 32768;
      if ( ctr1 < 2 ) ctr1 = 2;
      ctr2 = product / ctr1;
      error = abs(product - (long) ctr2 * (long) ctr1);

      while ( error && ctr1 < 32768 && ctr2 > 1) {
          ctr1++;
          ctr2 = product / ctr1;
          error = abs(product - (long) ctr2 * (long) ctr1);
      }
  }

  /* we can't have ctr2 equal to 1, or system hangs */
  if ( ctr2 == 1 ) {
      ctr2++;
      ctr1 /= 2;
  }

  board->ADC_ctr1 = ctr1;
  board->ADC_ctr2 = ctr2;
  board->ADC_freq = 10000000/((long)ctr1*(long)ctr2);

  #ifdef DEBUG
      printk("SetADCPacerFreq: Pacer Frequecy set to %ld\n", BoardData.ADC_freq);
  #endif

  return 0;
}

static int SetDACPacerFreq(BoardRec *board)
{
  WORD ctr1, ctr2;
  LONG product, error, error2;

  if ( board->DAC_freq == 0 ) {
    return -1;
  }

  /* divide 10MHz by frequency */
  product =  (long) 10000000 / board->DAC_freq;

  /* check for rounding error */
  error = abs(10000000 - product*board->DAC_freq);
  error2 = abs(10000000 - (product+1)*board->DAC_freq);
  if ( error2 < error ) product++;

  /* Now the job is to find two 16 bit numbers, that when multiplied
     together are approximately equal to product.  Start by setting
     one of them, ctr1 to 2 (minimum settable value) and increment until
     the error is minimized and ctr2 is less than 32768.

     NOTE: In Mode 2, a value of 1 is illegal! Therefore, crt1 and crt2
     can never be 1.

  */

  ctr1 = product / 32768;
  if ( ctr1 < 2 ) ctr1 = 2;
  ctr2 = product / ctr1;
  error = abs(product - (long) ctr2 * (long) ctr1);

  while ( error && ctr1 < 32768  && ctr2 > 1 ) {
       ctr1++;
       ctr2 = product / ctr1;
       error = abs(product - (long) ctr2 * (long) ctr1);
  }

  /* the frequency is prime, add 1 to it */
  if ( error ) {
      product++;
      ctr1 = product / 32768;
      if ( ctr1 < 2 ) ctr1 = 2;
      ctr2 = product / ctr1;
      error = abs(product - (long) ctr2 * (long) ctr1);

      while ( error && ctr1 < 32768 && ctr2 > 1) {
          ctr1++;
          ctr2 = product / ctr1;
          error = abs(product - (long) ctr2 * (long) ctr1);
      }
  }

  /* we can't have ctr2 equal to 1, or system hangs */
  if ( ctr2 == 1 ) {
      ctr2++;
      ctr1 /= 2;
  }

  board->DAC_ctr1 = ctr1;
  board->DAC_ctr2 = ctr2;
  board->DAC_freq = 10000000/((long)ctr1*(long)ctr2);

  #ifdef DEBUG
      printk("SetDACPacerFreq: Pacer Frequecy set to %ld\n", BoardData.DAC_freq);
  #endif

  return 0;
}

/***************************************************************************
 *
 * Load two part frequency to pacer counter chip.
 *
 ***************************************************************************/

static void LoadADCPacer(BoardRec *board)
{
  BYTE mask;
  BYTE bData;

  /* Write the values of ctr1 and ctr2 into counter A1 and A2 */

  #ifdef DEBUG
      printk("LoadADCPacer: freq = %ld   load values: ctr1 %#x ctr2 %#x\n", 
              board->ADC_freq, board->ADC_ctr1, board->ADC_ctr2);
  #endif

  mask = C2+MODE2+LSBFIRST;
  outb_p(mask, COUNTERA_CONTROL); 

  bData = (BYTE) (board->ADC_ctr2 & 0xff);
  outb_p(bData, COUNTERA_2_DATA); 

  bData = (BYTE) (board->ADC_ctr2 >> 8);
  outb_p(bData, COUNTERA_2_DATA); 

  mask = C1+MODE2+LSBFIRST;
  outb_p(mask, COUNTERA_CONTROL); 

  bData = (board->ADC_ctr1 & 0xff);
  outb_p(bData, COUNTERA_1_DATA); 

  bData = (board->ADC_ctr1 >> 8) & 0xff;
  outb_p(bData, COUNTERA_1_DATA); 
}

static void LoadDACPacer(BoardRec *board)
{
  BYTE mask;
  BYTE bData;

  /* Write the values of ctr1 and ctr2 into counter A1 and A2 */

  #ifdef DEBUG
      printk("LoadDACPacer: load values: ctr1 %#x ctr2 %#x\n", board->DAC_ctr1, board->DAC_ctr2);
  #endif

  mask = C2+MODE2+LSBFIRST;
  outb_p(mask, COUNTERB_CONTROL); 

  bData = (BYTE) (board->DAC_ctr2 & 0xff);
  outb_p(bData, COUNTERB_2_DATA); 

  bData = (BYTE) (board->DAC_ctr2 >> 8);
  outb_p(bData, COUNTERB_2_DATA); 

  mask = C1+MODE2+LSBFIRST;
  outb_p(mask, COUNTERB_CONTROL); 

  bData = (board->DAC_ctr1 & 0xff);
  outb_p(bData, COUNTERB_1_DATA); 

  bData = (board->DAC_ctr1 >> 8) & 0xff;
  outb_p(bData, COUNTERB_1_DATA); 
}

/***************************************************************************
 *
 * Load value into Counter 0  XXXX    Mode    MSB     LSB
 *                            Byte 3  Byte 2  Byte 1  Byte 0
 *
 ***************************************************************************/

static void LoadCounter0( LONG value )
{
  BYTE mask;
  BYTE bData;

  /* Write the value into Counter 0 Mode 2 */

  #ifdef DEBUG
      printk("LoadCounter0: load value %#lx into Counter 0.\n", value);
  #endif

  /* the mode is in the thrid byte */
  mask = 0xff & (value >> 16);
  mask += C0+LSBFIRST;
  outb_p(mask, COUNTERB_CONTROL); 

  if ( value & 0xff000000 ) {       /* load control word only */
      return;
  } else {
      /* LSB in the first byte of value */
      bData = (BYTE) (value & 0xff);
      outb_p(bData, COUNTERB_0_DATA); 

      /* MSB in the second byte of value */
      bData = (BYTE) ((value >> 8) & 0xff);
      outb_p(bData, COUNTERB_0_DATA); 
  }
}

/***************************************************************************
 *
 *  Load the residual sample count.  
 *
 *  Counter 0 is used to stop the acquisition when the desired number of samples 
 *  have been gathered.  It is gated on when a "residual" number of conversions
 *  remain.  Counter 0 will be enabled by use of the ARM bit.  Counter 0 is
 *  to be operated in Mode 0.
 *
 ***************************************************************************/

static void LoadResidualCount( WORD resCount )
{
  BYTE bReg;

  bReg = C0+MODE0+LSBFIRST;         /* 8254A Counter 0 -> mode 0 */
  outb_p(bReg, COUNTERA_CONTROL);

  /* Set word of total count */
  bReg = (BYTE) (resCount & 0xff);
  outb_p(bReg, COUNTERA_0_DATA);         
  bReg = (BYTE) (resCount >> 8);
  outb_p(bReg, COUNTERA_0_DATA);
  
  return;
}

/***************************************************************************
 *
 * Set which channels read() is interested in.
 *
 ***************************************************************************/

static int SetADCChannelMux( BYTE lowChan, BYTE highChan )
{
  BYTE channelMask;

  #ifdef DEBUG
    printk("SetADCChannelMux: lowChan = %d      highChan = %d\n", lowChan, highChan);
  #endif

  channelMask = (highChan << 4) | lowChan;
  BoardData.nSpare1 &= ~(0xff);
  BoardData.nSpare1 |= channelMask;
  return 0;
}

/***************************************************************************
 *
 * Set ADC gain
 *
 ***************************************************************************/

static int SetADCGain( WORD gain )
{
  #ifdef DEBUG
    printk("SetADCGain: gain = %#x\n", gain);
  #endif

 /* if the gain is different from the one already loaded,
    then calibration coeffs have to be loaded */

  if ( gain != (BoardData.nSpare1 & (UNIBIP | GS1 | GS0))) {
    calibrate_pci_das1602_ain(gain & ~SEDIFF);
  }
  
  BoardData.nSpare1 &= ~(UNIBIP | GS1 | GS0);
  BoardData.nSpare1 |= gain;
  return 0;
}

/***************************************************************************
 *
 * Turn on ADC pacer timer chip.
 *
 ***************************************************************************/

static void StartADCPacer()
{
  BoardData.nSpare1 &= ~ADPS1;
  BoardData.nSpare1 |=  ADPS0;
  outw_p(BoardData.nSpare1, MUX_REG);
  udelay(10);              /* wait 10 usecs for things to settle down */
  outw_p(BoardData.nSpare2 | TS0 | XTRCL, TRIG_REG);

  #ifdef DEBUG
      printk("StartADCPacer: MUX_REG = %#x\n", BoardData.nSpare1);
      printk("StartADCPacer: TRIG_REG = %#x\n", BoardData.nSpare2);
  #endif
}

/***************************************************************************
 *
 * Turn off ADC pacer timer chip.
 *
 ***************************************************************************/

static void StopADCPacer()
{
  BYTE mask;

  BoardData.nSpare1 &= ~(ADPS1 | ADPS0);
  outw_p(BoardData.nSpare1, MUX_REG);

  mask = C2+MODE2+LSBFIRST;
  outb_p(mask, COUNTERA_CONTROL);
  mask = C1+MODE2+LSBFIRST;
  outb_p(mask, COUNTERA_CONTROL);
}

/***************************************************************************
 *
 * Turn on DAC pacer timer chip.
 *
 ***************************************************************************/

static void StartDACPacer()
{
  BoardData.nSpare3 &= ~DAPS1;
  BoardData.nSpare1 |=  DAPS0;
  outw_p(BoardData.nSpare3, DAC_REG);

  #ifdef DEBUG
      printk("StartDACPacer: DAC_REG = %#x\n", BoardData.nSpare3);
  #endif
}

/***************************************************************************
 *
 * Turn off DAC pacer timer chip.
 *
 ***************************************************************************/

static void StopDACPacer()
{
  BYTE mask;

  BoardData.nSpare3 &= ~(DAPS1 | DAPS0);
  outw_p(BoardData.nSpare3, DAC_REG);

  mask = C2+MODE2+LSBFIRST;
  outb_p(mask, COUNTERB_CONTROL);
  mask = C1+MODE2+LSBFIRST;
  outb_p(mask, COUNTERB_CONTROL);
}

/***************************************************************************
 *
 * Handles 
 *    o pacer/counter triggered read() with software start.
 *
 * Pacer Method: 
 *
 *    o Pacer counter controls frequency of sample conversions
 *    o Total counter controls number of samples taken
 *    o Interrupts occur every 512(PACKETSIZE) samples and at last sample
 *    o Get chan->count samples
 *    o Samples are saved in kernel buffer(kernBuff) in readInterrupt()
 *
 ***************************************************************************/

static int PacerRead(WORD *kernBuff, ADC_ChanRec *chanRec)
{
  #ifdef DEBUG
      printk("Entering PacerRead().\n");
  #endif

  /* Prepare global data for parameterless interrupt handler */

  CurrChan = chanRec;                        /* pass chanel number to global */
  KernBuffPtr = KernBuff;                    /* same with KernBuff           */
  WordsToRead = chanRec->count;

  /* Disable everything first */
  outw(0x0, IRQ_REG);
  outw(0x0, TRIG_REG);

  /* Clear the AD FIFO */
  outw(0x1, ADC_FIFO_CLR);

  /* Load the residual count */
  LoadResidualCount(chanRec->count % PACKETSIZE);

  /* Set up the ADC Mux and Control register */
  SetADCChannelMux(chanRec->lowChan, chanRec->hiChan);  /* Read only this channel */
  SetADCGain(chanRec->gain);
 
  /* Load the Trigger Control/Status Register */
  if ( chanRec->count < 2*PACKETSIZE && chanRec->count >= PACKETSIZE) {
    BoardData.nSpare2 |= ARM;
    outw_p(BoardData.nSpare2, TRIG_REG);
  } else if ( chanRec->count < PACKETSIZE ) {
    BoardData.nSpare2 |= ARM | FFM0;
    outw_p(BoardData.nSpare2, TRIG_REG);
  }

  /* Enable interrupts */
  outw_p(DAHFCL | EOACL | INTCL | ADFLCL | DAEMCL, IRQ_REG);
  BoardData.nSpare0 = 0x0;
  BoardData.nSpare0 |= (INTE | INT1 | EOAIE);
  outw_p(BoardData.nSpare0, IRQ_REG);

  LoadADCPacer(&BoardData);                   /* Establish sample frequency */

  BoardData.nSpare1 &= ~ADPS1;               /* Select Pacer Source */
  BoardData.nSpare1 |=  ADPS0;
  outw_p(BoardData.nSpare1, MUX_REG);
  udelay(10);    /* wait 10 usecs for things to settle down */

  #ifdef DEBUG
    printk("PacerRead: Enter interrupt.  nSpare0 = %#x.\n", BoardData.nSpare0);
  #endif

  outw_p(BoardData.nSpare2 | TS0 | XTRCL, TRIG_REG);  /* let's go ... */
  interruptible_sleep_on(&adc0_wait);         /* Block in wait state */

  if ( WordsToRead != 0 ) {
      printk("Timing error in PacerRead: WordsToRead = %d\n", WordsToRead);
      return -1;
  }

  StopADCPacer();
  #ifdef DEBUG
    printk("Leaving PacerRead.\n");
  #endif
  return 0;
}


/***************************************************************************
 *
 * Handles 
 *    o pacer/counter triggered read() with external start.
 *    o external triggered read().
 *
 * Pacer Method: 
 *
 *    o Pacer counter controls frequency of sample conversions
 *    o Total counter controls number of samples taken
 *    o Interrupts occur every 512(PACKETSIZE) samples and at last sample
 *    o Get chan->count samples
 *    o Samples are saved in kernel buffer(kernBuff) in readInterrupt()
 *
 ***************************************************************************/

static int ExternRead(WORD *kernBuff, ADC_ChanRec *chanRec)
{

  #ifdef DEBUG
      printk("Entering ExternRead().\n");
  #endif

  /* Prepare global data for parameterless interrupt handler */

  CurrChan = chanRec;                        /* pass chanel number to global */
  KernBuffPtr = KernBuff;                    /* same with KernBuff           */
  WordsToRead = chanRec->count;

  /* Disable everything first */
  outw(0x0, IRQ_REG);
  outw(0x0, TRIG_REG);

  /* Clear the AD FIFO */
  outw(0x1, ADC_FIFO_CLR);

  /* Load the residual count */
  LoadResidualCount(chanRec->count % PACKETSIZE);

  /* Set up the ADC Mux and Control register */
  SetADCChannelMux(chanRec->lowChan, chanRec->hiChan);  /* Read only this channel */
  SetADCGain(chanRec->gain);
 
  /* Load the Trigger Control/Status Register */
  if ( chanRec->count < 2*PACKETSIZE && chanRec->count >= PACKETSIZE) {
    BoardData.nSpare2 |= ARM;
    outw_p(BoardData.nSpare2, TRIG_REG);
  } else if ( chanRec->count < PACKETSIZE ) {
    BoardData.nSpare2 |= ARM | FFM0;
    outw_p(BoardData.nSpare2, TRIG_REG);
  }

  /* Enable interrupts */
  outw_p(DAHFCL | EOACL | INTCL | ADFLCL | DAEMCL, IRQ_REG);
  BoardData.nSpare0 = 0x0;
  BoardData.nSpare0 = (INT1 | EOAIE);
  outw_p(BoardData.nSpare0, IRQ_REG);

  if (chanRec->mode == ADC_EXTERNAL_PACER_FALLING) {
    BoardData.nSpare1 &= ~ADPS0;               /* Select Pacer Source */
    BoardData.nSpare1 |=  ADPS1;               /* External Falling    */
  } else {
    BoardData.nSpare1 |=  ADPS0;               /* Select Pacer Source */
    BoardData.nSpare1 |=  ADPS1;               /* External Rising     */
  }

  outw_p(BoardData.nSpare1, MUX_REG);
  udelay(10);    /* wait 10 usecs for things to settle down */

  outw_p(BoardData.nSpare2 | TS0 | XTRCL, TRIG_REG);  /* let's go ... */
  interruptible_sleep_on(&adc0_wait);         /* Block in wait state */
  if ( WordsToRead != 0 ) {
      printk("Timing error in ExternRead: WordsToRead = %d\n", WordsToRead);
      return -1;
  }

  #ifdef DEBUG
    printk("Leaving ExternRead.\n");
  #endif

  return 0;
}


static void pci_das1602_aout( int channel, int gain, WORD value )
{
  WORD wReg = BoardData.nSpare3;

  /* Clear the DAC FIFO PTR4 BASE+2 */
  outw(0x0, DAC_FIFO_CLR);

  /* Set up DA pacer/mode/range at  PTR1 BASE+8 */
  wReg |= (LDAEMCL | DACEN);

  if ( channel ) {
    /* select channel 1 */
    wReg &= ~(HS0 | DAC1R0 | DAC1R1 );
    wReg |= HS1;
    switch( gain ) {
      case BP_5_00V:
        break;
      case BP_10_00V:
        wReg |=  DAC1R0;
        break;
      case UP_5_00V:
        wReg |=  DAC1R1;
        break;
      case UP_10_00V:
        wReg |=  (DAC1R0 | DAC1R1);
        break;
      default:
        printk("pci1602_AOut: unknown gain %d.\n", gain);
        return;
    }
  } else {
    /* select channel 0 */
    wReg &= ~(HS1 | DAC0R0 | DAC0R1 );
    wReg |= HS0;
    switch( gain ) {
      case BP_5_00V:
        break;
      case BP_10_00V:
        wReg |=  DAC0R0;
        break;
      case UP_5_00V:
        wReg |=  DAC0R1;
        break;
      case UP_10_00V:
        wReg |= (DAC0R0 | DAC0R1);
        break;
      default:
        printk("pci1602_AOut: unknown gain %d.\n", gain);
        return;
    }
  }
  
  outw(wReg, DAC_REG);

  /* 
     If the range has changed between the last output and the
     current one, then reload the calibration coefficients.
  */

  if ( channel ) {
    if ( (wReg & (DAC1R1 | DAC1R0)) != (BoardData.nSpare3 & (DAC1R1 | DAC1R0))) {
      calibrate_pci_das1602_aout( channel, gain );
      BoardData.nSpare3 &= ~(DAC1R1 | DAC1R0);                  /* clear the bits */
      BoardData.nSpare3 |= (wReg & (DAC1R1 | DAC1R0));          /* set the bits   */
    }
  } else {
    if ( (wReg & (DAC0R1 | DAC0R0)) != (BoardData.nSpare3 & (DAC0R1 | DAC0R0))) {
      calibrate_pci_das1602_aout( channel, gain );
      BoardData.nSpare3 &= ~(DAC0R1 | DAC0R0);                  /* clear the bits */
      BoardData.nSpare3 |= (wReg & (DAC0R1 | DAC0R0));          /* set the bits   */
    }
  }

  /* Output the value */

  outw(value, DAC_DATA_REG);
}

static void pci_das1602_aout_scan( int channel, int gain, WORD *DACbuf, int count )
{
  WORD wReg = 0x0;

  /* Disable everything first */
  outw(0x0, IRQ_REG);
  outw(0x0, DAC_REG);

  /* Clear the DAC FIFO */
  outw(0x0, DAC_FIFO_CLR);

  /* 
     If conversions are not done by external clock, set up
     the pacer clock.
  */

  if ( ChanDAC[channel].mode == DAC_PACER_CLOCK ) {
    LoadDACPacer(&BoardData);
  }

  /* Set up DA pacer/mode/range */
  switch( ChanDAC[channel].mode ) {
    case DAC_PACER_CLOCK:
      wReg |= DAPS0;
      break;
    case DAC_FALLING_EDGE:
      wReg |= DAPS1;
      break;
    case DAC_RISING_EDGE:
      wReg |= DAPS1;
      wReg |= DAPS0;
      break;
  }

  if ( channel & 0x1) {   /* channel 0 */
    wReg |= HS0;
    /* set up the gain */
    switch(  ChanDAC[0].gain) {
      case BP_10_00V:
        wReg |= DAC0R0;
        break;
      case UP_5_00V:
        wReg |= DAC0R1;
        break;
      case UP_10_00V:
        wReg |= DAC0R0 | DAC0R1;
        break;
    }
  }

  if ( channel & 0x2) {   /* channel 1 */
    wReg |= HS1;
    /* set up the gain */
    switch(  ChanDAC[1].gain) {
      case BP_10_00V:
        wReg |= DAC1R0;
        break;
      case UP_5_00V:
        wReg |= DAC1R1;
        break;
      case UP_10_00V:
        wReg |= DAC1R0 | DAC1R1;
        break;
    }
  }

  /* 
     If the range has changed between the last output
     and the current one, then reload the calibration
     coefficients.
   */

  if ( channel & 0x1) {  /* channel 0 */
    if ( (wReg & (DAC0R1 | DAC0R0)) != (BoardData.nSpare3 & (DAC0R1 | DAC0R0))) {
      calibrate_pci_das1602_aout( channel, gain );
      BoardData.nSpare3 &= ~(DAC0R1 | DAC0R0);                  /* clear the bits */
      BoardData.nSpare3 |= (wReg & (DAC0R1 | DAC0R0));          /* set the bits   */
    }
  }

  if ( channel & 0x2) {  /* channel 1 */
    if ( (wReg & (DAC1R1 | DAC1R0)) != (BoardData.nSpare3 & (DAC1R1 | DAC1R0))) {
      calibrate_pci_das1602_aout( channel, gain );
      BoardData.nSpare3 &= ~(DAC1R1 | DAC1R0);                  /* clear the bits */
      BoardData.nSpare3 |= (wReg & (DAC1R1 | DAC1R0));          /* set the bits   */
    }
  } 

  if ( count <= FIFO_SIZE ) {
    BoardData.nSpare0 |= DAEMIE;
    /* Load the DAC FIFO */
    RegCopyFromKernel(DAC_DATA_REG, (char*) DACbuf, count ); 
  } else {
    BoardData.nSpare0 |= DAHFIE;
    /* Load the DAC FIFO */
    RegCopyFromKernel(DAC_DATA_REG, (char *) DACbuf, FIFO_SIZE ); 
  }

  /* Write all the registers.  Enable interrupts. */
}

/*************************************************************************** 
 * Reads the calibration coeffs from non-volatile RAM in a PCI-DAS1602     *
 * and writes them to the appropriate trim-DACs associated with the analog * 
 * output sections.                                                        *
 * *************************************************************************/

void calibrate_pci_das1602_aout ( int channel, int gain )
{

  WORD addr;
  BYTE value;

  /*
     Get the starting offset in nvRAM at which the 
     calibration coeffs for the given range is stored
  */

  switch( gain ) {
    case BP_5_00V:
      addr = DAC_BIP5V_COEFF;
      break;
    case BP_10_00V:
      addr = DAC_BIP10V_COEFF;
      break;
    case UP_5_00V:
      addr = DAC_UNI5V_COEFF;
      break;
    case UP_10_00V:
      addr = DAC_UNI10V_COEFF;
      break;
    default:
      printk("pci1602_AOut: unknown gain %d.\n", gain);
      return;
  }

  if ( channel ) {
    addr += 0x10;  /* add 16 bytes if DAC1 */
    value = get_pci_das1602_nVRam( addr++ );
    write_pci_das1602_8800 ( DAC1_COARSE_GAIN, value );
    value = get_pci_das1602_nVRam( addr++ );
    write_pci_das1602_8800 ( DAC1_COARSE_OFFSET, value );
    value = get_pci_das1602_nVRam( addr++ );
    write_pci_das1602_8800 ( DAC1_FINE_GAIN, value );
    value = get_pci_das1602_nVRam( addr );
    write_pci_das1602_8800 ( DAC1_FINE_OFFSET, value );
  } else {
    /* DAC0 */
    value = get_pci_das1602_nVRam( addr++ );
    write_pci_das1602_8800 ( DAC0_COARSE_GAIN, value );
    value = get_pci_das1602_nVRam( addr++ );
    write_pci_das1602_8800 ( DAC0_COARSE_OFFSET, value );
    value = get_pci_das1602_nVRam( addr++ );
    write_pci_das1602_8800 ( DAC0_FINE_GAIN, value );
    value = get_pci_das1602_nVRam( addr );
    write_pci_das1602_8800 ( DAC0_FINE_OFFSET, value );
  }
}

/*************************************************************************** 
 * Reads the calibration coeffs from non-volatile RAM in a PCI-DAS1602     *
 * and writes them to the appropriate trim-DACs associated with the analog * 
 * input sections.                                                         *
 * *************************************************************************/

void calibrate_pci_das1602_ain( int gain )
{
  WORD addr;
  BYTE value;

  /*
     Get the starting offset in nvRAM at which the 
     calibration coeffs for the given range is stored.
  */

  #ifdef DEBUG
      printk("calibrate_pci_das1602_ain: resetting calibration coefficients. gain = %d\n",
              gain);
  #endif DEBUG

  switch( gain ) {
    case BP_10_00V:
      addr = ADC_BIP10V_COEFF;
      break;
    case BP_5_00V:
      addr = ADC_BIP5V_COEFF;
      break;
    case BP_2_50V:
      addr = ADC_BIP2PT5V_COEFF;
      break;
    case BP_1_25V:
      addr = ADC_BIP1PT25V_COEFF;
      break;
    case UP_10_00V:
      addr = ADC_UNI10V_COEFF;
      break;
    case UP_5_00V:
      addr = ADC_UNI5V_COEFF;
      break;
    case UP_2_50V:
      addr = ADC_UNI2PT5V_COEFF;
      break;
    case UP_1_25V:
      addr = ADC_UNI1PT25V_COEFF;
      break;
    default:
      printk("pci1602_ain: unknown gain %d.\n", gain);
      return;
  }

  value = get_pci_das1602_nVRam(addr++);  /* get front-end offset */
  write_pci_das1602_dac08(value);
  value = get_pci_das1602_nVRam(addr++);  /* get ADC offset */
  write_pci_das1602_8402(1, value);       /* pot #1 is for offset */
  value = get_pci_das1602_nVRam(addr);    /* get ADC gain */
  write_pci_das1602_8402(0, value);       /* pot #0 is for gain */

  udelay(1000);          /* waint 1 milliseconds when changing gains */
}


/*************************************************************************** 
 * Reads a byte from the specified address of the 
 * non-volative RAM on a PCI-DAS1602.
 * *************************************************************************/

BYTE get_pci_das1602_nVRam( WORD addr )
{
  /* See AMCC S5933 PCI Contrller Data Book, Spring 1996 */

  BYTE lo_addr;
  BYTE hi_addr;
  BYTE value;

  lo_addr = addr & 0xff;                /* low byte of address */
  hi_addr = (addr >> 0x8) & 0xff;       /* high byte of address */

  /* make sure device is not busy */
  while (inb(NVRAM_CTRL_REG) & 0x80);   /* D7=0 (not busy) */
  
  outb(0x80, NVRAM_CTRL_REG);           /* CMD to load low byte of address */
  outb(lo_addr, NVRAM_DATA_REG);        /* load low byte */

  outb(0xa0, NVRAM_CTRL_REG);           /* CMD to load high byte of address */
  outb(hi_addr, NVRAM_DATA_REG);        /* load low byte */

  outb(0xe0, NVRAM_CTRL_REG);           /* CMD to read NVRAM data */
  
  while (inb(NVRAM_CTRL_REG) & 0x80);   /* D7=0 (not busy) */
  value = inb(NVRAM_DATA_REG);
 
  return (value);
}

void write_pci_das1602_8800( BYTE addr, BYTE value )
{
  int i;

  /* write 3 bits MSB first of address, keep SEL8800 low */
  for ( i = 0; i < 3; i++) {
    if ( addr & 0x4 ) {
        outw_p(SDI, CALIBRATE_REG );
    } else {
        outw_p(0x0, CALIBRATE_REG );
    }
    addr <<= 1;
  }

  /* write 8 bits MSB first of data, keep SEL8800 low */
  for ( i = 0; i < 8; i++) {
    if ( value & 0x80 ) {
        outw_p(SDI, CALIBRATE_REG );
    } else {
        outw_p(0x0, CALIBRATE_REG );
    }
    value <<= 1;
  }

  /* SEL8800 to set load clock */
  outw_p( SEL8800, CALIBRATE_REG );

  /* SEL8800 to reset load clock and update output */
  outw_p( 0x0, CALIBRATE_REG );
}


void write_pci_das1602_8402( BYTE addr, BYTE value )
{
  int i;

  /* Select the device */
  outw_p( SEL8402, CALIBRATE_REG );

  /* Select one of the two potentiometers:
     0 = ADC gain pot
     1 = ADC offset pot
  */
  if ( addr ) {
    outw_p( SDI| SEL8402, CALIBRATE_REG );
  } else {
    outw_p( SEL8402, CALIBRATE_REG );
  }

  /* Write out each bit of the byte starting with the msb */
  for ( i = 0; i < 8; i++) {
    if ( value & 0x80 ) {
        outw_p(SDI | SEL8402, CALIBRATE_REG );
    } else {
        outw_p(SEL8402, CALIBRATE_REG );
    }
    value <<= 1;
  }

  /* Deselect the device */
  outw_p(0x0, CALIBRATE_REG );
}

void write_pci_das1602_dac08( BYTE value )
{
  /* Select front end offset value and load it */
  outw_p( SEL08, CALIBRATE_REG);
  outw_p( (WORD) value , CALIBRATE_REG);
}



